require 'base64'

module Ribbon::Intercom
  class Client
    class SDK
      autoload(:Adapters, 'ribbon/intercom/client/sdk/adapters')

      include Utils::Mixins::MockSafe

      attr_reader :adapter

      def initialize(*args)
        case args.first
        when Adapters::Adapter
          self.adapter = args.first
        else
          self.adapter = Adapters::HttpAdapter.new(*args)
        end
      end

      def begin
        Utils::MethodChain.begin { |methods|
          queue = Packet::MethodQueue.new
          methods.each { |meth, *args| queue.enqueue(meth, *args) }
          send_packet(method_queue: queue).retval
        }
      end

      ##
      # Returns the headers defined for this SDK. Optionally, you may also define
      # additional headers you'd like to add/override.
      def headers(new_headers={})
        adapter.headers(new_headers)
      end

      def connect(*args)
        adapter.connect(*args)
      end

      def connected?
        adapter.connected?
      end

      ##
      # Calls the method on the adapter returning the response Packet
      def call(method_name, *args)
        send_packet(
          method_queue: Packet::MethodQueue.new.enqueue(method_name, *args)
        )
      end

      ##
      # Sends the packet with the adapter, returning the response Packet.
      #
      # Intended to be called by Package.
      def send_packet(packet)
        _process_response(adapter.send_packet(packet))
      end

      def dup
        super.tap { |sdk|
          sdk.adapter = adapter.dup
        }
      end

      protected

      def adapter=(adapter)
        @adapter = adapter
      end

      private

      ##
      # Simulates calling the remote method as if it were a local method.
      #
      # Intended to be called by end-users.
      def method_missing(meth, *args, &block)
        call(meth, *args).retval
      end

      ##
      # Process a Packet object returned by Adapter#send_packet.
      def _process_response(packet)
        _handle_response_error(packet) if packet.error?

        Package.init_packages(packet.retval, self)
        packet
      end

      ##
      # Raises an error depending on what went wrong.
      def _handle_response_error(packet)
        raise 'packet contains no error' unless packet.error?
        raise packet.error
      end
    end # SDK
  end # Client
end # Ribbon::Intercom