require 'base64'

module Ribbon::Intercom
  class Package
    include Utils::Mixins::MockSafe

    class << self
      ##
      # Package up any non-basic objects that include Packageable::Mixin.
      def package(subject)
        Utils.walk(subject) { |subject, context|
          if Utils.basic_type?(subject)
            subject
          elsif context == :hash_key
            # Hash keys must be basic types.
            raise Errors::UnsafeResponseError, subject.inspect
          elsif subject.is_a?(Packageable::Mixin)
            _package_obj(subject, package(subject.package_data))
          elsif subject.is_a?(Class) && subject < Packageable::Mixin
            _package_obj(subject)
          else
            raise Errors::UnsafeResponseError, subject.inspect
          end
        }
      end

      def _package_obj(subject, data=nil)
        new(encode_subject(subject), data)
      end

      def encode_subject(subject)
        Marshal.dump(subject)
      end

      def decode_subject(encoded_subject)
        Marshal.load(encoded_subject)
      end

      ##
      # Walks the object and initializes all packages (i.e., sets them up to be
      # accessed by the end-user on the client).
      def init_packages(object, sdk)
        Utils.walk(object) { |object|
          object.send(:_init, sdk) if object.is_a?(Package)
          object
        }
      end
    end # Class Methods

    attr_reader :sdk

    def initialize(subject_data=nil, data=nil)
      @_subject_data = subject_data
      @_data = data
    end

    ##
    # Begin a method chain (must be completed with `#end`).
    #
    # When `#end` is called on the method chain, the chain will be resolved
    # remotely. This allows the chain of methods to be executed in one round-trip.
    def begin
      Utils::MethodChain.begin { |methods|
        queue = Packet::MethodQueue.new
        methods.each { |meth, *args| queue.enqueue(meth, *args) }
        _send_method_queue(queue)
      }
    end

    private

    def method_missing(meth, *args, &block)
      if @_data && @_data.key?(meth)
        @_data[meth]
      elsif sdk
        _call(meth, *args)
      else
        super
      end
    end

    ##
    # Call a method on the subject remotely.
    def _call(method_name, *args)
      _send_method_queue(Packet::MethodQueue.new.enqueue(method_name, *args))
    end

    ##
    # Initializes the package on the client-side.
    def _init(sdk)
      @sdk = sdk
      self.class.init_packages(@_data, sdk)
      mock_safe! if sdk.mock_safe?
    end

    def _send_method_queue(method_queue)
      _process_response(
        sdk.send_packet(
          subject:      @_subject_data,
          method_queue: method_queue
        )
      )
    end

    def marshal_dump
      [@_subject_data, @_data]
    end

    def marshal_load(array)
      @_subject_data = array[0]
      @_data = array[1]
    end

    def _process_response(packet)
      @_subject_data = packet.subject
      @_data = self.class.init_packages(packet.package_data, sdk)

      packet.retval
    end
  end # Package
end # Ribbon::Intercom