require 'securerandom'
require 'bcrypt'

module Ribbon::Intercom
  class Service
    class Channel
      autoload(:Stores, 'ribbon/intercom/service/channel/stores')

      include BCrypt

      attr_reader :name
      attr_reader :store
      attr_reader :token
      attr_reader :signing_keys
      attr_reader :secret_hash_crt
      attr_reader :secret_hash_prv

      def initialize(store, params={})
        @store = store
        @name = params[:name] or raise Errors::ChannelNameMissingError
        @token = params[:token]
        refresh(params)
      end

      ##
      # Refreshes the channel. To be called by the store after obtaining a lock.
      def refresh(params={})
        @secret_hash_crt = params[:secret_hash_crt]
        @secret_hash_prv = params[:secret_hash_prv]
        @signing_keys = params[:signing_keys] || {}

        @_allowed_to = nil
        may(*params.fetch(:may, []))
      end

      def rotate_secret
        SecureRandom.hex(16).tap { |secret|
          @secret_hash_prv = secret_hash_crt
          @secret_hash_crt = Password.create(secret)
        }
      end

      def rotate_secret!
        rotate_secret.tap { save }
      end

      def sign(data)
        key_id, key = _signing_key
        "\x01" + [key_id].pack('N') + Utils::Signer.new(key).sign(data)
      end

      def verify(signed_data)
        key_id = signed_data.slice(1, 4).unpack('N').first

        if (key=_retrieve_signing_key(key_id))
          Utils::Signer.new(key).verify(signed_data[5..-1])
        end
      end

      def may(*args)
        (@_allowed_to ||= Hash.new(false)).merge!(
          Hash[
            args.map { |perm| [perm.to_s, true] }
          ]
        ).keys.to_set
      end

      def may!(*args)
        may(*args).tap { save }
      end

      def may?(*perms)
        perms.all? { |perm| _may?(perm) }
      end

      def permissions
        @_allowed_to.keys.to_set
      end

      def valid_secret?(secret)
        !!secret && (secret_crt == secret || secret_prv == secret)
      end

      def secret_crt
        @__secret_crt ||= _to_bcrypt_pw(secret_hash_crt)
      end

      def secret_prv
        @__secret_prv ||= _to_bcrypt_pw(secret_hash_prv)
      end

      def with_lock(&block)
        store.with_lock(self, &block)
      end

      def save
        # Loop until unique
        unless token
          loop { break unless store.token_exists?(@token = SecureRandom.hex(4)) }
        end

        _run_validations
        store.persist(self)
      end

      def close
        store.delete(self)
      end

      def ==(other)
        other.is_a?(Channel) &&
          name == other.name &&
          store == other.store &&
          token == other.token &&
          secret_crt.to_s == other.secret_crt.to_s &&
          secret_prv.to_s == other.secret_prv.to_s &&
          permissions == other.permissions
      end

      private

      def _may?(required_perm)
        permissions.any? { |perm|
          regex = /\A#{Regexp.escape(perm).gsub('\*', '.*')}\z/
          regex.match(required_perm)
        }
      end

      def _to_bcrypt_pw(pw)
        Password.new(pw) if pw && !pw.empty?
      end

      def _run_validations
        raise Errors::ChannelNameMissingError unless name
        raise Errors::ChannelTokenMissingError unless token
        raise Errors::ChannelSecretMissingError unless secret_hash_crt
      end

      def _latest_signing_key_id
        ((signing_keys || {}).sort_by { |k, v| v[:timestamp] }.last || []).first
      end

      def _retrieve_signing_key(key_id)
        (data=_retrieve_signing_key_data(key_id)) && data[:key]
      end

      def _retrieve_signing_key_data(key_id)
        if key_id && (data=signing_keys[key_id])
          data.merge(key_id: key_id)
        else
          {}
        end
      end

      ##
      # Returns whether the key is expired. Optionally, an offset may be specified
      # to expire the key earlier or later.
      def _signing_key_expired?(key_id)
        _signing_key_ttl(key_id) <= 0
      end

      def _signing_key_ttl(key_id, ttl=3600)
        timestamp = _retrieve_signing_key_data(key_id)[:timestamp]
        (timestamp && (ttl - (Time.now.to_i - timestamp))).to_i
      end

      ##
      # Retrieve the latest signing key or generate a new one if the latest has
      # expired (older than 1 hour).
      def _signing_key
        latest_id = _latest_signing_key_id

        # Refresh the key if the current key has less than 5 minutes to live.
        if _signing_key_ttl(latest_id) < 300
          with_lock {
            latest_id = _latest_signing_key_id

            if _signing_key_ttl(latest_id) < 300
              latest_id = _add_signing_key
              _delete_expired_signing_keys
              save # Need to persist changes to signing keys.
            end
          }
        end

        [latest_id, _retrieve_signing_key(latest_id)]
      end

      def _add_signing_key
        (_latest_signing_key_id.to_i + 1).tap { |key_id|
          signing_keys[key_id] = {
            key: Utils::Signer.random_key,
            timestamp: Time.now.to_i
          }
        }
      end

      def _delete_expired_signing_keys
        signing_keys.reject! { |key_id| _signing_key_expired?(key_id) }
      end
    end # Service
  end # Channel
end # Ribbon::Intercom
