require 'redis'
require 'securerandom'
require 'base64'

module Ribbon::Intercom
  class Service
    module Channel::Stores
      class RedisStore < Store
        def initialize(params={})
          if params[:url]
            @_redis = Redis.new(url: params[:url])
          elsif params[:redis]
            @_redis = params[:redis]
          else
            raise Errors::InvalidStoreParamsError
          end
        end

        def token_exists?(token)
          @_redis.exists("channel:#{token}")
        end

        def lookup_channel(token)
          if (data=_load_data(token))
            Channel.new(self, data)
          end
        end

        def persist(channel)
          raise Errors::InvalidChannelError, channel.inspect unless channel.is_a?(Channel)

          channel.tap { |channel|
            data_hash = {}
            [:name, :token, :secret_hash_crt, :secret_hash_prv].each { |key|
              value = channel.send(key)
              data_hash[key] = value if value
            }

            # Save channel data as hash
            @_redis.mapped_hmset(_key_name(channel), data_hash)

            # Associate permissions set to its channel
            channel.permissions.each { |p|
              @_redis.sadd(_key_name(channel, 'permissions'), p)
            }

            # Save signing keys
            key_hash = Hash[
              channel.signing_keys.map { |key_id, data|
                [key_id, Base64.strict_encode64(Marshal.dump(data))]
              }
            ]

            @_redis.mapped_hmset(_key_name(channel, 'signing_keys'), key_hash) unless key_hash.empty?
          }
        end

        def delete(channel)
          raise Errors::InvalidChannelError, channel.inspect unless channel.is_a?(Channel)

          @_redis.del(_key_name(channel))
          @_redis.del(_key_name(channel, 'permissions'))
          nil
        end

        def with_lock(channel, ttl=1000)
          Mutex.new(@_redis, _key_name(channel, 'lock')).synchronize(ttl) { |validity|
            start_time = Mutex.time_in_ms
            channel.refresh(_load_data(channel.token))
            yield validity - (Mutex.time_in_ms - start_time)
          }
        end

        private

        def _load_data(token)
          channel_data_future = nil
          permissions_future  = nil
          signing_keys_future = nil

          @_redis.pipelined {
            channel_data_future = @_redis.hgetall(_key_name(token))
            permissions_future  = @_redis.smembers(_key_name(token, 'permissions'))
            signing_keys_future = @_redis.hgetall(_key_name(token, 'signing_keys'))
          }

          channel_data = channel_data_future.value

          if channel_data && !channel_data.empty?
            permissions = permissions_future.value

            signing_keys = signing_keys_future.value.map { |key_id, data|
              [key_id.to_i, Marshal.load(Base64.strict_decode64(data))]
            }.to_h

            Utils.symbolize_keys(
              channel_data.merge(
                may: permissions,
                signing_keys: signing_keys
              )
            )
          end
        end

        ##
        # A redis mutex inspired by <https://github.com/leandromoreira/redlock-rb>
        class Mutex
          class LockUnobtainableError < StandardError; end

          ##
          # Lua script to send to Redis when releasing a lock.
          # See: http://redis.io/commands/set
          UNLOCK_SCRIPT = <<-LUA
            if redis.call("get",KEYS[1]) == ARGV[1] then
              return redis.call("del",KEYS[1])
            else
              return 0
            end
          LUA

          DEFAULT_RETRY_COUNT = 3
          DEFAULT_RETRY_DELAY = 200
          CLOCK_DRIFT_FACTOR  = 0.01

          class << self
            def time_in_ms
              (Time.now.to_f * 1000).to_i
            end
          end # Class Methods

          attr_reader :redis
          attr_reader :name
          attr_reader :retry_count
          attr_reader :retry_delay

          def initialize(redis, name, opts={})
            @redis = redis
            @name = name
            @retry_count = opts[:retry_count] || DEFAULT_RETRY_COUNT
            @retry_delay = opts[:retry_delay] || DEFAULT_RETRY_DELAY
          end

          def synchronize(ttl)
            retry_count.times {
              start_time = self.class.time_in_ms

              if (handle=lock(ttl))
                begin
                  time_elapsed = (self.class.time_in_ms - start_time).to_i
                  return yield(ttl - time_elapsed - _drift(ttl))
                ensure
                  unlock(handle)
                end
              else
                sleep(rand(retry_delay).to_f / 1000)
              end
            }

            raise LockUnobtainableError, name
          end

          ##
          # Acquire a lock (non-blocking).
          def lock(ttl=1000)
            handle = SecureRandom.uuid
            redis.set(name, handle, nx: true, px: ttl) && handle
          end

          ##
          # Release a lock.
          def unlock(handle)
            redis.eval(UNLOCK_SCRIPT, [name], [handle])
          rescue
            # Do nothing, unlocking is best effort.
          end

          private

          ##
          # Approximate clock drift.
          def _drift(ttl)
            # Add 2 milliseconds to the drift to account for Redis expires
            # precision, which is 1 millisecond, plus 1 millisecond min drift
            # for small TTLs.
            (ttl * CLOCK_DRIFT_FACTOR).to_i + 2
          end
        end # Mutex

        def _key_name(channel_or_token, *sub_names)
          token = channel_or_token.is_a?(Channel) ? channel_or_token.token : channel_or_token
          ['channel', token, *sub_names].join(':')
        end
      end # RedisStore
    end # Channel::Stores
  end # Service
end # Ribbon::Intercom
