require 'securerandom'
require 'openssl'

module Ribbon::Intercom
  module Utils
    class Signer
      class << self
        def random_key
          SecureRandom.random_bytes(32)
        end

        def random_salt
          SecureRandom.random_bytes(8)
        end
      end

      attr_reader :key

      def initialize(key=self.class.random_key)
        raise ArgumentError, "key must be defined" unless key
        @key = key.dup.freeze
        @_digest = OpenSSL::Digest::SHA256.new
      end

      def sign(data)
        unless data.is_a?(String) && data.encoding == Encoding::BINARY
          raise ArgumentError, "data must be a binary encoded string"
        end

        salt = self.class.random_salt
        signature = _sign(salt, data)
        _encode(signature, salt, data)
      end

      def verify(signed_data)
        unless signed_data.is_a?(String) && signed_data.encoding == Encoding::BINARY
          raise ArgumentError, "signed_data must be a binary encoded string"
        end

        signature, salt, data = _decode(signed_data)
        data if _sign(salt, data) == signature
      end

      private

      def _sign(salt, data)
        OpenSSL::HMAC.digest(@_digest, key, salt + data)
      end

      def _encode(signature, salt, data)
        "\x01" + signature + salt + data
      end

      def _decode(signed_data)
        index = 0
        version = signed_data[index]

        index += version.length
        signature = signed_data.slice(index, @_digest.length)

        index += signature.length
        salt = signed_data.slice(index, 8)

        index += salt.length
        data = signed_data.slice(index..-1)

        [signature, salt, data]
      end
    end # Signer
  end # Utils
end # Ribbon::Intercom